@@ -1,11 +1,13 @@
 __author__ = "desultory"
-__version__ = "3.11.4"
+__version__ = "4.1.3"
 
 from pathlib import Path
+from shutil import rmtree, which
 from typing import Union
 
+from ugrd import AutodetectError, ValidationError
 from zenlib.types import NoDupFlatList
-from zenlib.util import colorize, contains
+from zenlib.util import colorize, contains, unset
 
 
 def detect_tmpdir(self) -> None:
@@ -17,11 +19,29 @@ def detect_tmpdir(self) -> None:
         self["tmpdir"] = Path(tmpdir)
 
 
+def _get_shell_path(self, shell_name) -> Path:
+    """Gets the real path to the shell binary."""
+    if shell := which(shell_name):
+        return Path(shell).resolve()
+    else:
+        raise AutodetectError(f"Shell '{shell_name}' not found.")
+
+
+def get_shell(self) -> None:
+    """Gets the shell, uses /bin/sh if a shell is not set"""
+    if shell := self["shell"]:
+        self.logger.info("Using shell: %s", colorize(shell, "blue", bright=True))
+        self["binaries"] = shell
+        shell_path = _get_shell_path(self, shell)
+        self["symlinks"]["shell"] = {"target": "/bin/sh", "source": shell_path}
+    else:
+        self.logger.info("Using default shell: %s", colorize("/bin/sh", "cyan"))
+        self["binaries"] = "/bin/sh"  # This should pull /bin/sh and the target if it's a symlink
+
+
 @contains("clean", "Skipping cleaning build directory", log_level=30)
 def clean_build_dir(self) -> None:
     """Cleans the build directory."""
-    from shutil import rmtree
-
     build_dir = self._get_build_path("/")
 
     if build_dir.is_dir():
@@ -37,17 +57,36 @@ def generate_structure(self) -> None:
         self._mkdir(subdir)
 
 
+def add_conditional_dependencies(self) -> None:
+    """Adds conditional dependencies to the dependencies list.
+    Keys are the dependency, values are a tuple of the condition type and value.
+    """
+
+    def add_dep(dep: str) -> None:
+        try:  # Try to add it as a binary, if it fails, add it as a dependency
+            self["binaries"] = dep
+        except (ValueError, ValidationError):
+            self["dependencies"] = dep
+
+    for dependency, condition in self["conditional_dependencies"].items():
+        condition_type, condition_value = condition
+        match condition_type:
+            case "contains":
+                contains(condition_value)(add_dep(dependency))
+            case "unset":
+                unset(condition_value)(add_dep(dependency))
+
+
 def calculate_dependencies(self, binary: str) -> list[Path]:
     """Calculates the dependencies of a binary using lddtree
     :param binary: The binary to calculate dependencies for
     :return: A list of dependency paths
     """
-    from shutil import which
     from subprocess import run
 
     binary_path = which(binary)
     if not binary_path:
-        raise RuntimeError("'%s' not found in PATH" % binary)
+        raise AutodetectError("'%s' not found in PATH" % binary)
 
     binary_path = Path(binary_path)
 
@@ -58,7 +97,7 @@ def calculate_dependencies(self, binary: str) -> list[Path]:
         self.logger.warning(
             "Unable to calculate dependencies for: %s" % colorize(binary, "red", bold=True, bright=True)
         )
-        raise RuntimeError("Unable to resolve dependencies, error: %s" % dependencies.stderr.decode("utf-8"))
+        raise AutodetectError("Unable to resolve dependencies, error: %s" % dependencies.stderr.decode("utf-8"))
 
     dependency_paths = []
     for dependency in dependencies.stdout.decode("utf-8").splitlines():
@@ -67,22 +106,31 @@ def calculate_dependencies(self, binary: str) -> list[Path]:
             dependency = dependency[1:]
 
         dependency_paths.append(Path(dependency))
-
+    self.logger.debug("[%s] Calculated dependencies: %s" % (binary, dependency_paths))
     return dependency_paths
 
 
+@contains("merge_usr", "Skipping /usr merge", log_level=30)
 def handle_usr_symlinks(self) -> None:
-    """Adds symlinks for /usr/bin and /usr/sbin to /bin and /sbin."""
-    build_dir = self._get_build_path("/")
+    """
+    Adds symlinks for /bin and /sbin to /usr/bin
+    Adds a symlink for /usr/sbin to /usr/bin (-> bin)
+    Adds smlinks for /lib to /usr/lib and /lib64 to /usr/lib64
+    Warns if the symlink path is a directory on the host system.
+    """
+    bin_symlink = ("bin", "usr/bin")
+    sbin_symlink = ("sbin", "usr/bin")
+    usr_sbin_symlink = ("usr/sbin", "bin")  # Make it relative
+    lib_symlink = ("lib", "usr/lib")
+    lib64_symlink = ("lib64", "usr/lib64")
+    symlinks = [bin_symlink, sbin_symlink, usr_sbin_symlink, lib_symlink, lib64_symlink]
 
-    if not (build_dir / "bin").is_dir():
-        if (build_dir / "usr/bin").is_dir():
-            self._symlink("/usr/bin", "/bin/")
-        else:
-            raise RuntimeError("Neither /bin nor /usr/bin exist in the build directory")
-
-    if not (build_dir / "sbin").is_dir() and (build_dir / "usr/sbin").is_dir():
-        self._symlink("/usr/sbin", "/sbin/")
+    for target, source in symlinks:
+        host_path = Path("/").joinpath(target)
+        if host_path.is_dir() and not host_path.is_symlink():
+            self.logger.warning("Host path is a directory: %s" % host_path)
+            self.logger.warning("Set `merge_usr = false` to disable /usr merge.")
+        self._symlink(source, target)
 
 
 def deploy_dependencies(self) -> None:
@@ -172,9 +220,16 @@ def find_libgcc(self) -> None:
     """
     from pathlib import Path
 
-    cmd = self._run(["ldconfig", "-p"], fail_silent=True, fail_hard=False)
+    musl_warning = False
+    try:
+        cmd = self._run(["ldconfig", "-p"], fail_silent=True, fail_hard=False)
+    except FileNotFoundError:
+        musl_warning = True
 
-    if b"Unimplemented option: -p" in cmd.stderr:  # Probably musl libc
+    if not musl_warning and b"Unimplemented option: -p" in cmd.stderr:  # Probably musl libc
+        musl_warning = True
+
+    if musl_warning:
         self.logger.warning("This check can be disabled by setting `find_libgcc = false` in the configuration.")
         return self.logger.warning("Unable to run ldconfig -p, if GCC is being used, this is fatal!")
 
@@ -206,8 +261,11 @@ def _process_out_file(self, out_file: str) -> None:
         return
 
     if "/" in out_file:  # If the out_file contains a path, resolve it
-        out_file = Path(out_file).resolve()
-        self.logger.info("Resolved relative output path: %s" % out_file)
+        out_file = Path(out_file)
+        resolved_out_file = out_file.resolve()
+        if resolved_out_file != out_file:
+            out_file = resolved_out_file
+            self.logger.info("Resolved relative output path: %s" % out_file)
     else:
         out_file = Path(out_file)
 
@@ -258,6 +316,7 @@ def _process_binaries_multi(self, binary: str) -> None:
     # The first dependency will be the path of the binary itself, don't add this to the library paths
     self["dependencies"] = dependencies[0]
     for dependency in dependencies[1:]:
+        self.logger.debug("[%s] Adding dependency: %s" % (binary, dependency))
         self["dependencies"] = dependency
         if str(dependency.parent) not in self["library_paths"]:
             self.logger.info("Adding library path: %s" % dependency.parent)
@@ -275,7 +334,7 @@ def _validate_dependency(self, dependency: Union[Path, str]) -> None:
         dependency = Path(dependency)
 
     if not dependency.exists():
-        raise FileNotFoundError("Dependency does not exist: %s" % dependency)
+        raise ValidationError("Dependency does not exist: %s" % dependency)
 
     return dependency
 
@@ -303,6 +362,10 @@ def _process_dependencies_multi(self, dependency: Union[Path, str]) -> None:
             self["symlinks"][f"_auto_{dependency.name}"] = {"source": resolved_path, "target": dependency}
             dependency = resolved_path
 
+    if dependency.is_symlink():
+        dependency = dependency.resolve()
+        self.logger.debug("Dependency target is a symlink, resolved to: %s" % dependency)
+
     self.logger.debug("Added dependency: %s" % dependency)
     self["dependencies"].append(dependency)
 
