@@ -1,6 +1,6 @@
-__version__ = "1.1.2"
+__version__ = "1.2.0"
 
-from zenlib.util import unset
+from zenlib.util import colorize, unset
 
 
 @unset("test_kernel")
@@ -8,7 +8,7 @@ def find_kernel_path(self):
     """Finds the kernel path for the current system"""
     from pathlib import Path
 
-    self.logger.info("Trying to find the kernel path for: %s", self["kernel_version"])
+    self.logger.info("Trying to find the kernel path for: %s", colorize(self["kernel_version"], "blue"))
     kernel_path = Path(self["_kmod_dir"]) / "vmlinuz"  # try this first
     if not (self["_kmod_dir"] / "vmlinuz").exists():
         for search_dir in ["/boot", "/efi"]:
@@ -21,7 +21,7 @@ def find_kernel_path(self):
         else:
             raise FileNotFoundError("Kernel not found: %s" % self["kernel_version"])
 
-    self.logger.info("Found kernel at: %s", kernel_path)
+    self.logger.info("Found kernel at: %s", colorize(kernel_path, "cyan"))
     self["test_kernel"] = kernel_path
 
 
@@ -37,10 +37,16 @@ def init_test_vars(self):
 def _get_qemu_cmd_args(self, test_image):
     """Returns arguements to run QEMU for the current test configuration."""
     test_initrd = self._get_out_path(self["out_file"])
-    self.logger.info("Testing initramfs image: %s", test_initrd)
+    self.logger.info("Testing initramfs image: %s", colorize(test_initrd, "yellow", bold=True))
     test_rootfs = test_image._get_out_path(test_image["out_file"])
-    self.logger.info("Test rootfs image: %s", test_rootfs)
-    self.logger.info("Test kernel: %s", self["test_kernel"])
+    self.logger.info(
+        "Test rootfs image: %s",
+        colorize(
+            test_rootfs,
+            "yellow",
+        ),
+    )
+    self.logger.info("Test kernel: %s", colorize(self["test_kernel"], "yellow"))
     qemu_args = {
         "-m": self["test_memory"],
         "-cpu": self["test_cpu"],
@@ -60,10 +66,12 @@ def _get_qemu_cmd_args(self, test_image):
 
     return arglist
 
+
 def get_copy_config_types(self) -> dict:
-    """ Returns the 'custom_parameters' name, type pairs for config to be copied to the test image """
+    """Returns the 'custom_parameters' name, type pairs for config to be copied to the test image"""
     return {key: self["custom_parameters"][key] for key in self["test_copy_config"] if key in self["custom_parameters"]}
 
+
 def make_test_image(self):
     """Creates a new initramfs generator to create the test image"""
     from ugrd.initramfs_generator import InitramfsGenerator
@@ -96,8 +104,8 @@ def test_image(self):
     qemu_cmd = _get_qemu_cmd_args(self, image)
 
     self.logger.debug("Test config:\n%s", image)
-    self.logger.info("Test flag: %s", self["test_flag"])
-    self.logger.info("QEMU command: %s", " ".join([str(arg) for arg in qemu_cmd]))
+    self.logger.info("Test flag: %s", colorize(self["test_flag"], "magenta"))
+    self.logger.info("QEMU command: %s", colorize(" ".join([str(arg) for arg in qemu_cmd]), bold=True))
 
     try:
         results = self._run(qemu_cmd, timeout=self["test_timeout"])
@@ -111,7 +119,7 @@ def test_image(self):
     for line in stdout:
         if line.endswith("exitcode=0x00000000"):
             panic_time = line.split("]")[0][1:].strip()
-            self.logger.info("Test took: %s", panic_time)
+            self.logger.info("Test took: %s", colorize(panic_time, "yellow", bold=True, bright=True))
             break
     else:
         self.logger.warning("Unable to determine test duration from panic message.")
