@@ -1,8 +1,9 @@
 __author__ = "desultory"
-__version__ = "3.9.1"
+__version__ = "4.0.3"
 
 from pathlib import Path
 
+from ugrd import AutodetectError, ValidationError
 from zenlib.util import colorize, contains, unset
 
 _module_name = "ugrd.crypto.cryptsetup"
@@ -51,7 +52,7 @@ def _process_cryptsetup_key_types_multi(self, key_type: str, config: dict) -> No
     else:
         # Make sure the key type has a key command
         if "key_command" not in config:
-            raise ValueError("Missing key_command for key type: %s" % key_type)
+            raise ValidationError("Missing key_command for key type: %s" % key_type)
         self["cryptsetup_key_types"][key_type] = config
 
 
@@ -97,7 +98,7 @@ def _validate_cryptsetup_config(self, mapped_name: str) -> None:
             )
     elif not any([config.get("partuuid"), config.get("uuid"), config.get("path")]):
         if not self["autodetect_root_luks"]:
-            raise ValueError(
+            raise ValidationError(
                 "A device uuid, partuuid, or path must be specified for cryptsetup mount: %s" % mapped_name
             )
 
@@ -130,12 +131,6 @@ def _process_cryptsetup_multi(self, mapped_name: str, config: dict) -> None:
         if config.get(f"include_{file_type}"):
             self["dependencies"] = config[f"{file_type}_file"]
 
-    if not config.get("retries"):
-        self.logger.info(
-            "[%s:%s] No retries specified, using default: %s" % (_module_name, mapped_name, self["cryptsetup_retries"])
-        )
-        config["retries"] = self["cryptsetup_retries"]
-
     self["cryptsetup"][mapped_name] = config
 
 
@@ -144,18 +139,24 @@ def _get_dm_info(self, mapped_name: str) -> dict:
     for device_info in self["_vblk_info"].values():
         if device_info["name"] == mapped_name:
             return device_info
-    raise KeyError("No device mapper information found for: %s" % mapped_name)
+    raise AutodetectError("No device mapper information found for: %s" % mapped_name)
 
 
 def _get_dm_slave_info(self, device_info: dict) -> (str, dict):
     """Gets the device mapper slave information for a particular device."""
     slave_source = device_info["slaves"][0]
-    try:
-        slave_device = f"/dev/{slave_source}"
-        return slave_device, self["_blkid_info"][slave_device]
-    except KeyError:
-        slave_device = f"/dev/mapper/{slave_source}"
-        return slave_device, self["_blkid_info"][slave_device]
+    slave_name = self["_vblk_info"].get(slave_source, {}).get("name")
+    search_paths = ["/dev/", "/dev/mapper/"]
+
+    for path in search_paths:
+        slave_device = path + slave_source
+        if slave_device in self["_blkid_info"]:
+            return slave_device, self["_blkid_info"][slave_device]
+        if slave_name:
+            slave_device = path + slave_name
+            if slave_device in self["_blkid_info"]:
+                return slave_device, self["_blkid_info"][slave_device]
+    raise AutodetectError("No slave device information found for: %s" % device_info)
 
 
 def _read_cryptsetup_header(self, mapped_name: str, slave_device: str = None) -> dict:
@@ -238,11 +239,11 @@ def _validate_cryptsetup_header(self, mapped_name: str) -> None:
 
     luks_info = _read_cryptsetup_header(self, mapped_name)
     if not luks_info:
-        raise ValueError("[%s] Unable to read LUKS header." % mapped_name)
+        raise ValidationError("[%s] Unable to read LUKS header." % mapped_name)
 
     if uuid := cryptsetup_info.get("uuid"):
         if luks_info.get("uuid") != uuid:
-            raise ValueError(
+            raise ValidationError(
                 "[%s] LUKS UUID mismatch, found '%s', expected: %s" % (mapped_name, luks_info["uuid"], uuid)
             )
 
@@ -252,13 +253,13 @@ def _validate_cryptsetup_header(self, mapped_name: str) -> None:
     if not self["argon2"]:  # if argon support was not detected, check if the header wants it
         for keyslot in luks_info.get("keyslots", {}).values():
             if keyslot.get("kdf", {}).get("type") == "argon2id":
-                raise FileNotFoundError("[%s] Missing cryptsetup dependency: libargon2.so" % mapped_name)
+                raise ValidationError("[%s] Missing cryptsetup dependency: libargon2.so" % mapped_name)
 
     if "header_file" in cryptsetup_info:
         self["check_included_or_mounted"] = cryptsetup_info["header_file"]  # Add the header file to the check list
 
 
-@contains("hostonly", "Skipping cryptsetup device check.", log_level=30)
+@contains("validate", "Skipping cryptsetup device check.", log_level=30)
 def _validate_cryptsetup_device(self, mapped_name) -> None:
     """Validates a cryptsetup device against the device mapper information,
     blkid information, and cryptsetup information.
@@ -277,7 +278,7 @@ def _validate_cryptsetup_device(self, mapped_name) -> None:
     for token_type in ["partuuid", "uuid"]:  # Validate the uuid/partuuid token against blkid info
         if cryptsetup_token := cryptsetup_info.get(token_type):
             if blkid_info.get(token_type) != cryptsetup_token:
-                raise ValueError(
+                raise ValidationError(
                     "[%s] LUKS %s mismatch, found '%s', expected: %s"
                     % (mapped_name, token_type, cryptsetup_token, blkid_info[token_type])
                 )
@@ -312,8 +313,16 @@ def detect_cryptsetup_backend(self) -> None:
 
 def _get_openssl_kdfs(self) -> list[str]:
     """Gets the available KDFs from OpenSSL"""
-    kdfs = self._run(["openssl", "list", "-kdf-algorithms"]).stdout.decode().lower().split("\n")
-    self.logger.debug("OpenSSL KDFs: %s" % kdfs)
+    kdfs = (
+        self._run(["openssl", "list", "-kdf-algorithms"], fail_hard=False, fail_silent=True)
+        .stdout.decode()
+        .lower()
+        .split("\n")
+    )
+    if not kdfs:
+        self.logger.warning("Unable to determine available OpenSSL KDFs.")
+    else:
+        self.logger.debug("OpenSSL KDFs: %s" % kdfs)
     return kdfs
 
 
@@ -331,7 +340,10 @@ def detect_argon2(self) -> None:
             if not any(dep.name.startswith("libcrypto.so") for dep in cryptsetup_deps):
                 self.logger.error("Cryptsetup is linked against OpenSSL, but libcrypto.so is not in dependencies.")
         case "gcrypt":
-            gcrypt_version = self._run(["libgcrypt-config", "--version"]).stdout.decode().strip().split("-")[0]
+            try:
+                gcrypt_version = self._run(["libgcrypt-config", "--version"]).stdout.decode().strip().split("-")[0]
+            except RuntimeError as e:
+                raise AutodetectError("Unable to determine Libgcrypt version: %s" % e)
             maj, minor, patch = map(int, gcrypt_version.split("."))
             if not any(dep.name.startswith("libgcrypt.so") for dep in cryptsetup_deps):
                 self.logger.error("Cryptsetup is linked against Libgcrypt, but libgcrypt.so is not in dependencies.")
@@ -383,7 +395,9 @@ def export_crypt_sources(self) -> list[str]:
     Adds the cryptsetup source and token to the exports.
     Sets the token to the partuuid or uuid if it exists.
     Sets the SOURCE when using a path.
-    Only allows using the path if validation is disabled."""
+    Only allows using the path if validation is disabled.
+    sets CRYPTSETUP_HEADER_{name} if a header file is set.
+    """
     for name, parameters in self["cryptsetup"].items():
         _validate_luks_config(self, name)
         if parameters.get("path"):
@@ -394,9 +408,9 @@ def export_crypt_sources(self) -> list[str]:
                 self["exports"]["CRYPTSETUP_SOURCE_%s" % name] = parameters["path"]
                 self.logger.info("Set CRYPTSETUP_SOURCE_%s: %s" % (name, parameters.get("path")))
                 continue
-            raise ValueError("Validation must be disabled to use device paths with the cryptsetup module.")
+            raise ValidationError("Validation must be disabled to use device paths with the cryptsetup module.")
         elif not parameters.get("partuuid") and not parameters.get("uuid") and parameters.get("path"):
-            raise ValueError("Device source for cryptsetup mount must be specified: %s" % name)
+            raise ValidationError("Device source for cryptsetup mount must be specified: %s" % name)
 
         for token_type in ["partuuid", "uuid"]:
             if token := parameters.get(token_type):
@@ -404,112 +418,167 @@ def export_crypt_sources(self) -> list[str]:
                 self.logger.debug("Set CRYPTSETUP_TOKEN_%s: %s=%s" % (name, token_type.upper(), token))
                 break
         else:
-            raise ValueError("A partuuid or uuid must be specified for cryptsetup mount: %s" % name)
+            raise ValidationError("A partuuid or uuid must be specified for cryptsetup mount: %s" % name)
+        if header_file := parameters.get("header_file"):
+            self["exports"]["CRYPTSETUP_HEADER_%s" % name] = header_file
+            self.logger.debug("Set CRYPTSETUP_HEADER_%s: %s" % (name, header_file))
 
 
-def get_crypt_dev(self) -> list[str]:
+def get_crypt_dev(self) -> str:
     """Gets the device path for a particular cryptsetup device at runtime.
     First attempts to read CRYPTSETUP_SOURCE_{name} if it exists.
     If it doesn't exist, or the device is not found, it will attempt to resolve the device using the token.
     If that doesn't exist, it will fail."""
-    return [
-        'source_dev="$(readvar CRYPTSETUP_SOURCE_"$1")"',
-        'source_token="$(readvar CRYPTSETUP_TOKEN_"$1")"',
-        'if [ -n "$source_dev" ]; then',
-        '    if [ -e "$source_dev" ]; then',
-        '        echo -n "$source_dev"',
-        "        return",
-        "    fi",
-        "fi",
-        'if [ -n "$source_token" ]; then',
-        '    source_dev=$(blkid --match-token "$source_token" --output device)',
-        '    if [ -n "$source_dev" ]; then',
-        '        echo -n "$source_dev"',
-        "    fi",
-        "fi",
+    return """
+    source_dev="$(readvar CRYPTSETUP_SOURCE_"$1")"
+    source_token="$(readvar CRYPTSETUP_TOKEN_"$1")"
+    if [ -n "$source_dev" ]; then
+        if [ -e "$source_dev" ]; then
+            printf "%s" "$source_dev"
+            return
+        fi
+    fi
+    if [ -n "$source_token" ]; then
+        source_dev=$(blkid --match-token "$source_token" --output device)
+        if [ -n "$source_dev" ]; then
+            printf "%s" "$source_dev"
+        fi
+    fi
+    if [ -z "$source_dev" ]; then
+        rd_fail "Failed to find cryptsetup device: $1"
+    fi
+    """
+
+
+def open_crypt_dev(self) -> str:
+    """Returns a shell script to open a cryptsetup device.
+    The first argument is the device name, for get_crypt_dev, and as the mapped name
+    The second argument is a key file, if it exists.
+        This key file may be a named pipe, previously created by the key_command.
+    """
+    out = """
+    crypt_device="$(get_crypt_dev "$1")"
+    header="$(readvar CRYPTSETUP_HEADER_"$1")"
+
+    cryptsetup_args="cryptsetup open --tries 1"
+    if [ -n "$header" ]; then
+        einfo "[$1] Opening cryptsetup device with detached header: $header"
+        cryptsetup_args="$cryptsetup_args --header $header"
+    fi
+    if [ -e /run/vars/key_data ]; then
+        einfo "[$1] Unlocking with key data from: /run/vars/key_data"
+        cryptsetup_args="$cryptsetup_args --key-file /run/vars/key_data"
+    elif [ -n "$2" ]; then
+        einfo "[$1] Unlocking with key file: $2"
+        cryptsetup_args="$cryptsetup_args --key-file $2"
+    fi
+    """
+    if self["cryptsetup_trim"]:
+        out += 'cryptsetup_args="$cryptsetup_args --allow-discards"'
+    return (
+        out
+        + """
+    cryptsetup_args="$cryptsetup_args $crypt_device $1"
+    edebug "[$1] cryptsetup args: $cryptsetup_args"
+    $cryptsetup_args
+    """
+    )
+
+
+def _open_crypt_dev(self, name: str, parameters: dict) -> list[str]:
+    """Generates a loop to open a cryptsetup device with the given parameters."""
+    from textwrap import dedent
+
+    retries = parameters.get("retries", self["cryptsetup_retries"])
+    out = [
+        f"einfo 'Opening cryptsetup device: {name}'",
+        f'retries={retries}; i=0; while [ "$((i=i+1))" -le $retries ]; do',
     ]
 
+    key_file = parameters.get("key_file")
+    reset_command = parameters.get("reset_command")
+    reset_lines = [reset_command, "continue"] if reset_command else ["continue"]
+    key_command = parameters.get("key_command")
+    plymouth_key_command = parameters.get("plymouth_key_command") if "ugrd.base.plymouth" in self["modules"] else None
 
-def open_crypt_device(self, name: str, parameters: dict) -> list[str]:
-    """Returns a bash script to open a cryptsetup device."""
-    self.logger.debug("[%s] Processing cryptsetup volume: %s" % (name, parameters))
-    retries = parameters["retries"]
+    _key_command_lines = f"""    einfo "($i/$retries)[{name}] Running key command: {key_command}"
+    if ! {key_command} > /run/vars/key_data; then
+        ewarn 'Failed to run key command: {key_command}'
+        {reset_command}
+        continue
+    fi"""
 
-    out = [f"prompt_user 'Press enter to unlock device: {name}'"] if self["cryptsetup_prompt"] else []
-    out += [f"for ((i = 1; i <= {retries}; i++)); do"]
-
-    # Resolve the device source using get_crypt_dev, if no source is returned, run rd_fail
-    out += [
-        f'    crypt_dev="$(get_crypt_dev {name})"',
-        '    if [ -z "$crypt_dev" ]; then',
-        f'        rd_fail "Failed to resolve device source for cryptsetup mount: {name}"',
-        "    fi",
-    ]
-
-    reset_command = parameters.get("reset_command", "continue")
-    # When there is a key command, evaluate it into $key_data
-    if "key_command" in parameters:
-        self.logger.debug("[%s] Using key command: %s" % (name, parameters["key_command"]))
-        out += [
-            f"    einfo 'Attempting to open LUKS key: {parameters['key_file']}'",
-            f"    edebug 'Using key command: {parameters['key_command']}'",
-        ]
-
-        if "plymouth_key_command" in parameters and "ugrd.base.plymouth" in self["modules"]:
+    if key_command:
+        self.logger.debug("[%s] Using key command: %s" % (name, key_command))
+        out += ["    rm -f /run/vars/key_data"]  # Remove the key data file if it exists
+        if plymouth_key_command:
+            self.logger.debug("[%s] Using plymouth key command: %s" % (name, plymouth_key_command))
             out += [
                 "    if plymouth --ping; then",
-                f'        plymouth ask-for-password --prompt "[${{i}} / {retries}] Enter passphrase to unlock key for: {name}" --command "{parameters["plymouth_key_command"]}" --number-of-tries 1 > /run/vars/key_data || {reset_command}',
+                f'        if ! plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock key for: {name}" --command "{parameters["plymouth_key_command"]}" --number-of-tries 1 > /run/vars/key_data; then',
+                *[f"            {line}" for line in reset_lines],
+                "        fi",
                 "    else",
-                f'        {parameters["key_command"]} > /run/vars/key_data || {reset_command}',
+                *[f"        {line}" for line in dedent(_key_command_lines).split("\n")],
                 "    fi",
             ]
         else:
-            out += [f"    {parameters['key_command']} > /run/vars/key_data || {reset_command}"]
-
-    cryptsetup_command = "cryptsetup open --tries 1"  # Set tries to 1 since it runs in the loop
-    cryptsetup_target = f'"$crypt_dev" {name}'  # Add a variable for the source device and mapped name
-
-    if header_file := parameters.get("header_file"):  # Use the header file if it exists
-        out += [f"    einfo 'Using header file: {header_file}'"]
-        cryptsetup_command += f" --header {header_file}"
-
-    if self["cryptsetup_trim"]:
-        cryptsetup_command += " --allow-discards"
-        self.logger.warning("Using --allow-discards can be a security risk.")
-
-    # Check if the device was successfully opened
-    out += [
-        '    einfo "Unlocking device: $crypt_dev"',  # Unlock using key data if it exists
-        "    if [ -e /run/vars/key_data ]; then",
-        f"        if {cryptsetup_command} --key-file=/run/vars/key_data {cryptsetup_target}; then",
-        "            rm /run/vars/key_data",  # Remove the key data file
-        "            break",
-        "        fi",  # Try to open the device using plymouth if it's running
-        "        rm /run/vars/key_data",  # Remove the key data file
-    ]
-    if "ugrd.base.plymouth" in self["modules"]:  # Attempt plain unlock if using plymouth
+            out += [f"        {line}" for line in dedent(_key_command_lines).split("\n")]
+        out += [  # open_crypt_dev will use the key data file if it exists
+            f'    einfo "($i/$retries)[{name}] Unlocked key to var: key_data"',
+            f"    if open_crypt_dev {name}; then",
+            "        rm -f /run/vars/key_data",  # Remove the key data file after opening the device
+        ]
+    elif key_file:  # For plain key files, just use it as arg 2 for open_crypt_dev
         out += [
-            f'    elif plymouth --ping && plymouth ask-for-password --prompt "[${{i}} / {retries}] Enter passphrase to unlock {name}" --command "{cryptsetup_command} {cryptsetup_target}" --number-of-tries 1; then',
-            "        break",
-        ]  # Break if the device was successfully opened
-    if "key_file" in parameters:  # try a key file directly if it exists
-        out += [f'    elif {cryptsetup_command} --key-file {parameters["key_file"]} {cryptsetup_target}; then']
-    else:  # Otherwise, open directly
-        out += [f"    elif {cryptsetup_command} {cryptsetup_target}; then"]
-    out += ["        break", "    fi", f'    ewarn "Failed to open device: {name} ($i / {retries})"']
-    # Halt if the autoretry is disabled
+            f'    einfo "($i/$retries)[{name}] Using key file: {key_file}"',
+            f"    if open_crypt_dev {name} {key_file}; then",
+        ]
+    else:
+        if "ugrd.base.plymouth" in self["modules"]:
+            out += [  # When plymouth is used, write the entered key to the key data file, where it will be used by open_crypt_dev
+                "    if plymouth --ping; then",
+                f'        plymouth ask-for-password --prompt "[${{i}} / ${{retries}}] Enter passphrase to unlock: {name}" > /run/vars/key_data',
+                "    fi",
+                f'    einfo "($i/$retries) Unlocking device: {name}"',
+                f"    if open_crypt_dev {name}; then",
+            ]
+        else:
+            out += [f'    einfo "($i/$retries) Unlocking device: {name}"', f"    if open_crypt_dev {name}; then"]
+    # Break on success
+    out += ["        break", "    fi", f'    ewarn "($i/$retries) Failed to open cryptsetup device: {name}"']
     if not self["cryptsetup_autoretry"]:
-        out += ['    prompt_user "Press enter to retry"']
-    # Add the reset command if it exists
-    if reset_command != "continue":
+        out += ["    prompt_user 'Press enter to retry'"]
+    if reset_command:
         out += ['    einfo "Running key reset command"', f"    {reset_command}"]
-    out += ["done\n"]
+
+    out += [
+        "done",
+        "if [ -e /run/vars/key_data ]; then",
+        "    eerror 'Removing leftover key data file'",
+        "    rm -f /run/vars/key_data",
+        "fi",
+    ]
+
+    if parameters.get("try_nokey") and key_file:
+        new_params = parameters.copy()
+        for parameter in ["key_file", "key_command", "reset_command"]:
+            try:
+                new_params.pop(parameter)
+            except KeyError:
+                pass
+        out += [
+            f"if ! cryptsetup status {name} > /dev/null 2>&1; then",
+            f'    ewarn "[{name}] Failed to open device using key: {key_file}"',
+            *[f"    {sh_line}" for sh_line in _open_crypt_dev(self, name, new_params)],
+            "fi",
+        ]
 
     return out
 
 
 def crypt_init(self) -> list[str]:
-    """Generates the bash script portion to prompt for keys."""
+    """Generates the shell script portion to prompt for keys."""
     if self["loglevel"] > 5:
         self.logger.warning("loglevel > 5, cryptsetup prompts may not be visible.")
 
@@ -517,27 +586,11 @@ def crypt_init(self) -> list[str]:
     for name, parameters in self["cryptsetup"].items():
         # Check if the volume is already open, if so, skip it
         out += [
-            f"if cryptsetup status {name} > /dev/null 2>&1; then",
+            f"if ! cryptsetup status {name} > /dev/null 2>&1; then",
+            *[f"    {sh_line}" for sh_line in _open_crypt_dev(self, name, parameters)],
+            "else",
             f'    ewarn "Device already open: {name}"',
-            "    return",
             "fi",
-        ]
-        out += open_crypt_device(self, name, parameters)
-        if "try_nokey" in parameters and parameters.get("key_file"):
-            new_params = parameters.copy()
-            for parameter in ["key_file", "key_command", "reset_command"]:
-                try:
-                    new_params.pop(parameter)
-                except KeyError:
-                    pass
-            out += [
-                f"if ! cryptsetup status {name} > /dev/null 2>&1; then",
-                f'    ewarn "Failed to open device using keys: {name}"',
-            ]
-            out += [f"    {bash_line}" for bash_line in open_crypt_device(self, name, new_params)]
-            out += ["fi"]
-        # Check that the device was successfully opened
-        out += [
             f"if ! cryptsetup status {name} > /dev/null 2>&1; then",
             f'    rd_fail "Failed to open cryptsetup device: {name}"',
             "fi",
