@@ -1,9 +1,10 @@
 __author__ = "desultory"
-__version__ = "5.8.0"
+__version__ = "6.5.0"
 
 from pathlib import Path
 from typing import Union
 
+from ugrd import AutodetectError, ValidationError
 from zenlib.util import colorize, contains, pretty_print
 
 BLKID_FIELDS = ["uuid", "partuuid", "label", "type"]
@@ -21,33 +22,83 @@ MOUNT_PARAMETERS = [
 ]
 
 
-@contains("validate", "Skipping mount validation, validation is disabled.", log_level=30)
-def _validate_mount_config(self, mount_name: str, mount_config) -> None:
-    """Validate the mount config."""
-    if mount_config.get("no_validate"):
-        return self.logger.warning("Skipping mount validation: %s" % colorize(mount_name, "yellow", bold=True))
+def _get_device_id(device: str) -> str:
+    """Gets the device id from the device path."""
+    return Path(device).stat().st_rdev >> 8, Path(device).stat().st_rdev & 0xFF
 
-    for source_type in SOURCE_TYPES:
-        if source_type in mount_config:
-            self.logger.debug("[%s] Validated source type: %s" % (mount_name, mount_config))
-            break
-    else:  # If no source type is found, raise an error, unless it's the root mount
-        if source_type not in mount_config and mount_name != "root":
-            raise ValueError("[%s] No source type found in mount: %s" % (mount_name, mount_config))
 
-    for parameter, value in mount_config.copy().items():
-        self.logger.debug("[%s] Validating parameter: %s" % (mount_name, parameter))
-        if parameter == "options" and not mount_config.get("no_validate_options"):
-            for option in value:
-                if "subvol=" in option:
-                    if mount_name == "root":
-                        raise ValueError(
-                            "Please use the root_subvol parameter instead of setting the option manually in the root mount."
-                        )
-                    elif mount_config["type"] not in ["btrfs", "bcachefs"]:
-                        raise ValueError("subvol option can only be used with btrfs or bcachefs mounts.")
-        elif parameter not in MOUNT_PARAMETERS:
-            raise ValueError("Invalid parameter in mount: %s" % parameter)
+def _resolve_dev(self, device_path) -> str:
+    """Resolves a device to one indexed in blkid.
+
+    Takes the device path, such as /dev/root, and resolves it to a device indexed in blkid.
+    If the device is an overlayfs, resolves the lowerdir device.
+    """
+    if str(device_path) in self["_blkid_info"]:
+        self.logger.debug("Device already resolved to blkid indexed device: %s" % device_path)
+        return device_path
+
+    self.logger.debug("Resolving device: %s" % device_path)
+    mountpoint = _resolve_device_mountpoint(self, device_path)
+    device_path = _resolve_overlay_lower_device(self, mountpoint)
+    mountpoint = _resolve_device_mountpoint(self, device_path)  # May have changed if it was an overlayfs
+
+    major, minor = _get_device_id(self["_mounts"][mountpoint]["device"])
+    for device in self["_blkid_info"]:
+        check_major, check_minor = _get_device_id(device)
+        if (major, minor) == (check_major, check_minor):
+            self.logger.info("Resolved device: %s -> %s" % (device_path, colorize(device, "cyan")))
+            return device
+    self.logger.critical("Failed to resolve device: %s" % colorize(device_path, "red", bold=True))
+    self.logger.error("Blkid info: %s" % pretty_print(self["_blkid_info"]))
+    self.logger.error("Mount info: %s" % pretty_print(self["_mounts"]))
+    return device_path
+
+
+def _find_mountpoint(self, path: str) -> str:
+    """Finds the mountpoint of a file or directory,
+    Checks if the parent dir is a mountpoint, if not, recursively checks the parent dir."""
+    check_path = Path(path).resolve()
+    parent = check_path.parent if not check_path.is_dir() else check_path
+    if str(parent) in self["_mounts"]:
+        return str(parent)
+    elif parent == Path("/"):  # The root mount SHOULD always be found...
+        raise AutodetectError("Mountpoint not found for: %s" % path)
+    return _find_mountpoint(self, parent.parent)
+
+
+def _resolve_device_mountpoint(self, device) -> str:
+    """Gets the mountpoint of a device based on the device path."""
+    for mountpoint, mount_info in self["_mounts"].items():
+        if str(device) == mount_info["device"]:
+            return mountpoint
+    raise AutodetectError("Device mountpoint not found: %s" % device)
+
+
+def _resolve_overlay_lower_dir(self, mountpoint) -> str:
+    for option in self["_mounts"][mountpoint]["options"]:
+        if option.startswith("lowerdir="):
+            return option.removeprefix("lowerdir=")
+    raise AutodetectError(
+        "[%s] No lower overlayfs mountpoint found: %s" % mountpoint, self["_mounts"][mountpoint]["options"]
+    )
+
+
+def _resolve_overlay_lower_device(self, mountpoint) -> dict:
+    """Returns device for the lower overlayfs mountpoint.
+    If it's not an overlayfs, returns the device for the mountpoint.
+
+    If it is, iterate through the lowerdir devices until a non-overlayfs mount is found.
+    """
+    if self["_mounts"][mountpoint]["fstype"] != "overlay":
+        return self["_mounts"][mountpoint]["device"]
+
+    while self["_mounts"][mountpoint]["fstype"] == "overlay":
+        lowerdir = _resolve_overlay_lower_dir(self, mountpoint)
+        mountpoint = _find_mountpoint(self, lowerdir)
+        if mountpoint == "/":  # The lowerdir mount should never be the root mount
+            raise AutodetectError("Lowerdir mount not found: %s" % lowerdir)
+
+    return self["_mounts"][mountpoint]["device"]
 
 
 def _merge_mounts(self, mount_name: str, mount_config, mount_class) -> None:
@@ -68,6 +119,35 @@ def _merge_mounts(self, mount_name: str, mount_config, mount_class) -> None:
     return dict(self[mount_class][mount_name], **mount_config)
 
 
+@contains("validate", "Skipping mount validation, validation is disabled.", log_level=30)
+def _validate_mount_config(self, mount_name: str, mount_config) -> None:
+    """Validate the mount config."""
+    if mount_config.get("no_validate"):
+        return self.logger.warning("Skipping mount validation: %s" % colorize(mount_name, "yellow", bold=True))
+
+    for source_type in SOURCE_TYPES:
+        if source_type in mount_config:
+            self.logger.debug("[%s] Validated source type: %s" % (mount_name, mount_config))
+            break
+    else:  # If no source type is found, raise an error, unless it's the root mount
+        if source_type not in mount_config and mount_name != "root":
+            raise ValidationError("[%s] No source type found in mount: %s" % (mount_name, mount_config))
+
+    for parameter, value in mount_config.copy().items():
+        self.logger.debug("[%s] Validating parameter: %s" % (mount_name, parameter))
+        if parameter == "options" and not mount_config.get("no_validate_options"):
+            for option in value:
+                if "subvol=" in option:
+                    if mount_name == "root":
+                        raise ValueError(
+                            "Please use the root_subvol parameter instead of setting the option manually in the root mount."
+                        )
+                    elif mount_config["type"] not in ["btrfs", "bcachefs"]:
+                        raise ValueError("subvol option can only be used with btrfs or bcachefs mounts.")
+        elif parameter not in MOUNT_PARAMETERS:
+            raise ValueError("Invalid parameter in mount: %s" % parameter)
+
+
 def _process_mount(self, mount_name: str, mount_config, mount_class="mounts") -> None:
     """Processes the passed mount config."""
     mount_config = _merge_mounts(self, mount_name, mount_config, mount_class)
@@ -82,10 +162,14 @@ def _process_mount(self, mount_name: str, mount_config, mount_class="mounts") ->
 
     # Add imports based on the mount type
     if mount_type := mount_config.get("type"):
-        if mount_type in ["vfat", "ext4", "xfs"]:
+        if mount_type in ["vfat", "xfs"]:
             self["kmod_init"] = mount_type
         elif mount_type == "nilfs2":
             self["binaries"] = "mount.nilfs2"
+        elif mount_type == "ext4":
+            if "ugrd.fs.ext4" not in self["modules"]:
+                self.logger.info("Auto-enabling module: %s", colorize("ext4", "cyan"))
+                self["modules"] = "ugrd.fs.ext4"
         elif mount_type == "btrfs":
             if "ugrd.fs.btrfs" not in self["modules"]:
                 self.logger.info("Auto-enabling module: %s", colorize("btrfs", "cyan"))
@@ -105,16 +189,6 @@ def _process_mount(self, mount_name: str, mount_config, mount_class="mounts") ->
         self["paths"] = mount_config["destination"]
 
 
-def _process_mount_timeout(self, timeout: float) -> None:
-    """Set the mount timeout, enables mount_wait."""
-    if not isinstance(timeout, (int, float)):
-        raise ValueError("Invalid timeout: %s" % timeout)
-    if not self["mount_wait"]:
-        self.logger.info("Enabling mount wait, as a timeout is set: %s" % timeout)
-        self["mount_wait"] = True
-    self.data["mount_timeout"] = timeout
-
-
 def _process_mounts_multi(self, mount_name: str, mount_config) -> None:
     _process_mount(self, mount_name, mount_config)
 
@@ -228,6 +302,8 @@ def get_mounts_info(self) -> None:
             device, mountpoint, fstype, options, _, _ = line.split()
             self["_mounts"][mountpoint] = {"device": device, "fstype": fstype, "options": options.split(",")}
 
+    self.logger.debug("Mount info: %s" % pretty_print(self["_mounts"]))
+
 
 @contains("hostonly", "Skipping blkid enumeration, hostonly mode is disabled.", log_level=30)
 def get_blkid_info(self, device=None) -> dict:
@@ -236,13 +312,17 @@ def get_blkid_info(self, device=None) -> dict:
     The info is stored in self['_blkid_info']."""
     from re import search
 
-    if device:
-        blkid_output = self._run(["blkid", device]).stdout.decode().strip()
-    else:
-        blkid_output = self._run(["blkid"]).stdout.decode().strip()
+    try:
+        if device:
+            blkid_output = self._run(["blkid", device]).stdout.decode().strip()
+        else:
+            blkid_output = self._run(["blkid"]).stdout.decode().strip()
+    except RuntimeError:
+        self.logger.error("Blkid output: %s" % blkid_output)
+        raise AutodetectError("Failed to get blkid info for: %s" % device)
 
     if not blkid_output:
-        raise ValueError("Unable to get blkid info.")
+        raise AutodetectError("Unable to get blkid info.")
 
     for device_info in blkid_output.split("\n"):
         dev, info = device_info.split(": ")
@@ -261,32 +341,34 @@ def get_blkid_info(self, device=None) -> dict:
 
 
 @contains("init_target", "init_target must be set", raise_exception=True)
-@contains(
-    "autodetect_init_mount", "Skipping init mount autodetection, autodetect_init_mount is disabled.", log_level=30
-)
+@contains("autodetect_init_mount", "Init mount autodetection disabled, skipping.", log_level=30)
 @contains("hostonly", "Skipping init mount autodetection, hostonly mode is disabled.", log_level=30)
-def autodetect_init_mount(self, parent=None) -> None:
+def autodetect_init_mount(self) -> None:
     """Checks the parent directories of init_target, if the path is a mountpoint, add it to late_mounts."""
-    if not parent:
-        parent = self["init_target"].parent
-    if parent == Path("/"):
+    init_mount = _find_mountpoint(self, self["init_target"])
+    if init_mount == "/":
         return
-    if str(parent) in self["_mounts"]:
-        self.logger.info("Detected init mount: %s" % colorize(parent, "cyan"))
-        mount_name = str(parent).removeprefix("/")
-        mount_dest = str(parent)
-        mount_device = self["_mounts"][str(parent)]["device"]
-        mount_type = self["_mounts"][str(parent)]["fstype"]
-        mount_options = self["_mounts"][str(parent)]["options"]
-        blkid_info = self["_blkid_info"][mount_device]
-        mount_source_type, mount_source = _get_mount_source_type(self, blkid_info, with_val=True)
-        self["late_mounts"][mount_name] = {
-            "destination": mount_dest,
-            mount_source_type: mount_source,
-            "type": mount_type,
-            "options": mount_options,
-        }
-    autodetect_init_mount(self, parent.parent)
+
+    if init_mount in self["late_mounts"]:
+        return self.logger.debug("Init mount already detected: %s" % init_mount)
+
+    if init_mount not in self["_mounts"]:
+        raise AutodetectError("Init mount not found in host mounts: %s" % init_mount)
+
+    self.logger.info("Detected init mount: %s" % colorize(init_mount, "cyan"))
+    mount_name = init_mount.removeprefix("/")
+    mount_dest = init_mount
+    mount_device = self["_mounts"][init_mount]["device"]
+    mount_type = self["_mounts"][init_mount]["fstype"]
+    mount_options = self["_mounts"][init_mount]["options"]
+    blkid_info = self["_blkid_info"][mount_device]
+    mount_source_type, mount_source = _get_mount_source_type(self, blkid_info, with_val=True)
+    self["late_mounts"][mount_name] = {
+        "destination": mount_dest,
+        mount_source_type: mount_source,
+        "type": mount_type,
+        "options": mount_options,
+    }
 
 
 @contains("hostonly", "Skipping virtual block device enumeration, hostonly mode is disabled.", log_level=30)
@@ -318,7 +400,7 @@ def get_virtual_block_info(self) -> dict:
                 self["_vblk_info"][virt_device.name]["uuid"] = (virt_device / "md/uuid").read_text().strip()
                 self["_vblk_info"][virt_device.name]["level"] = (virt_device / "md/level").read_text().strip()
             else:
-                raise ValueError("Failed to get virtual device name: %s" % virt_device.name)
+                raise AutodetectError("Failed to get virtual device name: %s" % virt_device.name)
 
             try:
                 self["_vblk_info"][virt_device.name]["name"] = (virt_device / "dm/name").read_text().strip()
@@ -335,11 +417,6 @@ def get_virtual_block_info(self) -> dict:
         self.logger.debug("No virtual block devices found.")
 
 
-def _get_device_id(device: str) -> str:
-    """Gets the device id from the device path."""
-    return Path(device).stat().st_rdev >> 8, Path(device).stat().st_rdev & 0xFF
-
-
 @contains("hostonly", "Skipping device mapper autodetection, hostonly mode is disabled.", log_level=30)
 def _autodetect_dm(self, mountpoint, device=None) -> None:
     """Autodetects device mapper config given a mountpoint.
@@ -356,7 +433,7 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
     elif mountpoint:
         source_device = _resolve_overlay_lower_device(self, mountpoint)
     else:
-        raise FileNotFoundError("Mountpoint not found in host mounts: %s" % mountpoint)
+        raise AutodetectError("Mountpoint not found in host mounts: %s" % mountpoint)
 
     device_name = source_device.split("/")[-1]
     if not any(device_name.startswith(prefix) for prefix in ["dm-", "md"]):
@@ -372,9 +449,9 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
             elif f"/dev/mapper/{source_name}" in self["_blkid_info"]:
                 source_device = f"/dev/mapper/{source_name}"
             elif not get_blkid_info(self, source_device):
-                raise FileNotFoundError("[%s] No blkid info for virtual device: %s" % (mountpoint, source_device))
+                raise AutodetectError("[%s] No blkid info for virtual device: %s" % (mountpoint, source_device))
         else:
-            raise ValueError("[%s] No blkid info for virtual device: %s" % (mountpoint, source_device))
+            raise AutodetectError("[%s] No blkid info for virtual device: %s" % (mountpoint, source_device))
 
     self.logger.info("[%s] Detected virtual block device: %s" % (mountpoint, colorize(source_device, "cyan")))
     source_device = Path(source_device)
@@ -386,12 +463,12 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
             dev_name = name
             break
     else:
-        raise RuntimeError(
+        raise AutodetectError(
             "[%s] Unable to find device mapper device with maj: %s min: %s" % (source_device, major, minor)
         )
 
     if len(self["_vblk_info"][dev_name]["slaves"]) == 0:
-        raise RuntimeError("No slaves found for device mapper device, unknown type: %s" % source_device.name)
+        raise AutodetectError("No slaves found for device mapper device, unknown type: %s" % source_device.name)
     slave_source = self["_vblk_info"][dev_name]["slaves"][0]
 
     try:
@@ -400,9 +477,9 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
         if slave_source in self["_vblk_info"]:
             blkid_info = self["_blkid_info"][f"/dev/mapper/{self['_vblk_info'][slave_source]['name']}"]
         else:
-            raise KeyError("Unable to find blkid info for device mapper slave: %s" % slave_source)
+            raise AutodetectError("Unable to find blkid info for device mapper slave: %s" % slave_source)
     if source_device.name != self["_vblk_info"][dev_name]["name"] and source_device.name != dev_name:
-        raise ValueError(
+        raise ValidationError(
             "Device mapper device name mismatch: %s != %s" % (source_device.name, self["_vblk_info"][dev_name]["name"])
         )
 
@@ -421,8 +498,8 @@ def _autodetect_dm(self, mountpoint, device=None) -> None:
             self.logger.error(
                 "If LUKS headers are detached, they must be configured with the corresponding mapped device name."
             )
-            raise RuntimeError("[%s] No type found for device mapper device: %s" % (dev_name, source_device))
-        raise RuntimeError("Unknown device mapper device type: %s" % blkid_info.get("type"))
+            raise AutodetectError("[%s] No type found for device mapper device: %s" % (dev_name, source_device))
+        raise ValidationError("Unknown device mapper device type: %s" % blkid_info.get("type"))
 
     autodetect_mount_kmods(self, slave_source)
 
@@ -450,7 +527,7 @@ def autodetect_raid(self, mount_loc, dm_name, blkid_info) -> None:
         self.logger.info("[%s] MDRAID level: %s" % (mount_loc.name, colorize(level, "cyan")))
         self["_kmod_auto"] = level
     else:
-        raise ValueError("[%s] Failed to autodetect MDRAID level: %s" % (dm_name, blkid_info))
+        raise AutodetectError("[%s] Failed to autodetect MDRAID level: %s" % (dm_name, blkid_info))
 
 
 @contains("autodetect_root_lvm", "Skipping LVM autodetection, autodetect_root_lvm is disabled.", log_level=20)
@@ -465,7 +542,7 @@ def autodetect_lvm(self, mount_loc, dm_num, blkid_info) -> None:
         self.logger.info("[%s] LVM volume contianer uuid: %s" % (mount_loc.name, colorize(uuid, "cyan")))
         self["lvm"] = {self["_vblk_info"][dm_num]["name"]: {"uuid": uuid}}
     else:
-        raise ValueError("Failed to autodetect LVM volume uuid: %s" % mount_loc.name)
+        raise AutodetectError("Failed to autodetect LVM volume uuid: %s" % mount_loc.name)
 
 
 @contains("autodetect_root_luks", "Skipping LUKS autodetection, autodetect_root_luks is disabled.", log_level=30)
@@ -489,15 +566,15 @@ def autodetect_luks(self, mount_loc, dm_num, blkid_info) -> None:
 
     if len(self["_vblk_info"][dm_num]["slaves"]) > 1:
         self.logger.error("Device mapper slaves: %s" % colorize(self["_vblk_info"][dm_num]["slaves"], "red", bold=True))
-        raise RuntimeError("Multiple slaves found for device mapper device, unknown type: %s" % mount_loc.name)
+        raise AutodetectError("Multiple slaves found for device mapper device, unknown type: %s" % mount_loc.name)
 
     dm_type = blkid_info.get("type")
     if dm_type != "crypto_LUKS":
         if not blkid_info.get("uuid"):  # No uuid will be defined if there are detached headers
             if not self["cryptsetup"][mount_loc.name].get("header_file"):
-                raise ValueError("[%s] Unknown LUKS mount type: %s" % (mount_loc.name, dm_type))
+                raise AutodetectError("[%s] Unknown LUKS mount type: %s" % (mount_loc.name, dm_type))
         else:  # If there is some uuid and it's not LUKS, that's a problem
-            raise RuntimeError(
+            raise AutodetectError(
                 "[%s] Unknown device mapper slave type: %s" % (self["_vblk_info"][dm_num]["slaves"][0], dm_type)
             )
 
@@ -515,73 +592,45 @@ def autodetect_luks(self, mount_loc, dm_num, blkid_info) -> None:
     )
 
 
-def _resolve_dev(self, device_path) -> str:
-    """Resolves a device path, if possible.
-    Useful for cases where the device in blkid differs from the device in /proc/mounts.
-    """
-    major, minor = _get_device_id(self["_mounts"][device_path]["device"])
-    for device in self["_blkid_info"]:
-        check_major, check_minor = _get_device_id(device)
-        if (major, minor) == (check_major, check_minor):
-            self.logger.info(
-                "Resolved device: %s -> %s" % (self["_mounts"][device_path]["device"], colorize(device, "cyan"))
-            )
-            return device
-    self.logger.warning("Failed to resolve device: %s" % colorize(self["_mounts"]["/"]["device"], "red", bold=True))
-    return self["_mounts"][device_path]["device"]
-
-
-def _resolve_overlay_lower_dir(self, mountpoint) -> str:
-    for option in self["_mounts"][mountpoint]["options"]:
-        if option.startswith("lowerdir="):
-            return option.removeprefix("lowerdir=")
-    raise ValueError(
-        "[%s] No lower overlayfs mountpoint found: %s" % mountpoint, self["_mounts"][mountpoint]["options"]
-    )
-
-
-def _resolve_overlay_lower_device(self, mountpoint) -> dict:
-    """Returns device for the lower overlayfs mountpoint."""
-    if self["_mounts"][mountpoint]["fstype"] != "overlay":
-        return self["_mounts"][mountpoint]["device"]
-
-    lowerdir = _resolve_overlay_lower_dir(self, mountpoint)
-    return self["_mounts"][lowerdir]["device"]
-
-
 @contains("autodetect_root", "Skipping root autodetection, autodetect_root is disabled.", log_level=30)
 @contains("hostonly", "Skipping root autodetection, hostonly mode is disabled.", log_level=30)
 def autodetect_root(self) -> None:
     """Sets self['mounts']['root']'s source based on the host mount."""
     if "/" not in self["_mounts"]:
-        raise FileNotFoundError(
+        self.logger.error("Host mounts: %s" % pretty_print(self["_mounts"]))
+        raise AutodetectError(
             "Root mount not found in host mounts.\nCurrent mounts: %s" % pretty_print(self["_mounts"])
         )
     # Sometimes the root device listed in '/proc/mounts' differs from the blkid info
-    root_dev = self["_mounts"]["/"]["device"]
-    if self["resolve_root_dev"]:  # Sometimes the root device listed in '/proc/mounts' differs from the blkid info
-        root_dev = _resolve_dev(self, "/")
+    root_dev = _resolve_dev(self, self["_mounts"]["/"]["device"])
     if ":" in root_dev:  # only use the first device
         root_dev = root_dev.split(":")[0]
         for alt_devices in root_dev.split(":")[1:]:  # But ensure kmods are loaded for all devices
             autodetect_mount_kmods(self, alt_devices)
     _autodetect_mount(self, "/")
     if self["autodetect_root_dm"]:
-        _autodetect_dm(self, "/")
+        if self["mounts"]["root"]["type"] == "btrfs":
+            from ugrd.fs.btrfs import _get_btrfs_mount_devices
+
+            for device in _get_btrfs_mount_devices(self, "/", root_dev):
+                _autodetect_dm(self, "/", device)
+        else:
+            _autodetect_dm(self, "/")
 
 
 def _autodetect_mount(self, mountpoint) -> None:
     """Sets mount config for the specified mountpoint."""
     if mountpoint not in self["_mounts"]:
-        raise FileNotFoundError("auto_mount mountpoint not found in host mounts: %s" % mountpoint)
+        self.logger.error("Host mounts: %s" % pretty_print(self["_mounts"]))
+        raise AutodetectError("auto_mount mountpoint not found in host mounts: %s" % mountpoint)
 
-    mount_device = _resolve_overlay_lower_device(self, mountpoint)
+    mount_device = _resolve_overlay_lower_device(self, mountpoint)  # Just resolve the overlayfs device
 
     if ":" in mount_device:  # Handle bcachefs
         mount_device = mount_device.split(":")[0]
 
-    if mount_device not in self["_blkid_info"]:
-        get_blkid_info(self, mount_device)
+    if mount_device not in self["_blkid_info"]:  # Try to get the blkid info
+        get_blkid_info(self, mount_device)  # Raises an exception if it fails
 
     mount_info = self["_blkid_info"][mount_device]
     autodetect_mount_kmods(self, mount_device)
@@ -606,7 +655,7 @@ def _autodetect_mount(self, mountpoint) -> None:
             mount_config[mount_name][source_type] = source
             break
     else:
-        raise ValueError("[%s] Failed to autodetect mount source." % mountpoint)
+        raise AutodetectError("[%s] Failed to autodetect mount source." % mountpoint)
 
     self["mounts"] = mount_config
 
@@ -664,32 +713,29 @@ def mount_late(self) -> list[str]:
 
 def mount_fstab(self) -> list[str]:
     """Generates the init function for mounting the fstab.
-    If a mount_timeout is set, sets the default rootdelay.
-    If a mount_wait is set, enables rootwait.
-    mount_retries sets the number of times to retry the mount (for unattended booting).
+    Keeps re-attempting with mount_timeout or rootdelay until successful.
+    mount_retries sets the number of times to retry the mount, infinite otherwise.
     """
-    out = []
-    if timeout := self.get("mount_timeout"):  # Set the timeout, using the defined timeout as the default
-        out.append(f"timeout=$(readvar rootdelay {timeout})")
-    else:
-        out.append("timeout=$(readvar rootdelay)")
-
-    if rootwait := self.get("mount_wait"):  # Set the rootwait bool, using the defined rootwait as the default
-        out.append(f"rootwait=$(readvar rootwait {int(rootwait)})")
-    else:
-        out.append("rootwait=$(readvar rootwait)")
-
-    out += [
-        'if [ -z "$timeout" ]; then',  # If timeout is not set, prompt the user -
-        '    if [ "$rootwait" == "1" ]; then',  # only if rootwait is set to 1
-        '        prompt_user "Press enter once devices have settled."',
-        "    fi",
-        "else",  # If timeout is set, prompt the user with a timeout
-        '    prompt_user "Press enter once devices have settled. [${timeout}s]" "$timeout"',
-        "fi",
-        f'retry {self["mount_retries"]} "${{timeout:-1}}" mount -a || rd_fail "Failed to mount all filesystems."',
+    out = [
+        'einfo "Attempting to mount all filesystems."',
+        f"timeout=$(readvar rootdelay {self.get('mount_timeout', 1)})",
     ]
 
+    if retries := self.get("mount_retries"):
+        out += [
+            f'retry {retries} "$timeout" mount -a || rd_fail "Failed to mount all filesystems."',
+        ]
+    else:
+        out += [
+            "while ! mount -a; do",  # Actually retry forever, retry with a short timeout may fail
+            '    if prompt_user "Press enter to break, waiting: ${timeout}s" "$timeout"; then',
+            '        rd_fail "Failed to mount all filesystems."',
+            "    fi",
+            '    eerror "Failed to mount all filesystems, retrying."',
+            "done",
+            "einfo 'All filesystems mounted.'",
+        ]
+
     return out
 
 
@@ -697,7 +743,6 @@ def _validate_host_mount(self, mount, destination_path=None) -> bool:
     """Checks if a defined mount exists on the host."""
     if mount.get("no_validate"):
         return self.logger.warning("Skipping host mount validation for config:\n%s" % pretty_print(mount))
-
     if mount.get("base_mount"):
         return self.logger.debug("Skipping host mount validation for base mount: %s" % mount)
 
@@ -705,12 +750,9 @@ def _validate_host_mount(self, mount, destination_path=None) -> bool:
     # If a destination path is passed, like for /, use that instead of the mount's destination
     destination_path = str(mount["destination"]) if destination_path is None else destination_path
 
-    # Using the mount path, get relevant host mount info
-    host_source_dev = _resolve_overlay_lower_device(self, destination_path)
+    host_source_dev = _resolve_dev(self, self["_mounts"][destination_path]["device"])
     if ":" in host_source_dev:  # Handle bcachefs
         host_source_dev = host_source_dev.split(":")[0]
-    if destination_path == "/" and self["resolve_root_dev"]:
-        host_source_dev = _resolve_dev(self, "/")
 
     host_mount_options = self["_mounts"][destination_path]["options"]
     for option in mount.get("options", []):
@@ -719,7 +761,7 @@ def _validate_host_mount(self, mount, destination_path=None) -> bool:
         if option == "ro":  # Allow the ro option to be set in the config
             continue
         if option not in host_mount_options:
-            raise ValueError(
+            raise ValidationError(
                 "Host mount options mismatch. Expected: %s, Found: %s" % (mount["options"], host_mount_options)
             )
 
@@ -727,19 +769,19 @@ def _validate_host_mount(self, mount, destination_path=None) -> bool:
         if mount_val == Path(host_source_dev) or mount_val == host_source_dev:
             self.logger.debug("[%s] Host mount validated: %s" % (destination_path, mount))
             return True
-        raise ValueError(
+        raise ValidationError(
             "Host mount path device path does not match config. Expected: %s, Found: %s" % (mount_val, host_source_dev)
         )
     elif mount_type in ["uuid", "partuuid", "label"]:
         # For uuid, partuuid, and label types, check that the source matches the host mount
         if self["_blkid_info"][host_source_dev][mount_type] != mount_val:
-            raise ValueError(
+            raise ValidationError(
                 "Host mount source device mismatch. Expected: %s: %s, Found: %s"
                 % (mount_type, mount_val, host_source_dev)
             )
         self.logger.debug("[%s] Host mount validated: %s" % (destination_path, mount))
         return True
-    raise ValueError("[%s] Unable to validate host mount: %s" % (destination_path, mount))
+    raise ValidationError("[%s] Unable to validate host mount: %s" % (destination_path, mount))
 
 
 @contains("validate", "Skipping host mount validation, validation is disabled.", log_level=30)
@@ -751,16 +793,16 @@ def check_mounts(self) -> None:
         _validate_host_mount(self, mount, "/" if mount_name == "root" else None)
 
 
-def mount_root(self) -> list[str]:
+def mount_root(self) -> str:
     """Mounts the root partition to $MOUNTS_ROOT_TARGET."""
-    return [
-        'if grep -qs "$(readvar MOUNTS_ROOT_TARGET)" /proc/mounts; then',
-        '    ewarn "Root mount already exists, unmounting: $(readvar MOUNTS_ROOT_TARGET)"',
-        '    umount "$(readvar MOUNTS_ROOT_TARGET)"',
-        "fi",
-        '''einfo "Mounting '$(readvar MOUNTS_ROOT_SOURCE)' ($(readvar MOUNTS_ROOT_TYPE)) to '$(readvar MOUNTS_ROOT_TARGET)' with options: $(readvar MOUNTS_ROOT_OPTIONS)"''',
-        f'retry {self["mount_retries"]} {self["mount_timeout"] or 1} mount "$(readvar MOUNTS_ROOT_SOURCE)" -t "$(readvar MOUNTS_ROOT_TYPE)" "$(readvar MOUNTS_ROOT_TARGET)" -o "$(readvar MOUNTS_ROOT_OPTIONS)"',
-    ]
+    return f"""
+    if grep -qs "$(readvar MOUNTS_ROOT_TARGET)" /proc/mounts; then
+        ewarn "Root mount already exists, unmounting: $(readvar MOUNTS_ROOT_TARGET)"
+        umount "$(readvar MOUNTS_ROOT_TARGET)"
+    fi
+    einfo "Mounting '$(readvar MOUNTS_ROOT_SOURCE)' ($(readvar MOUNTS_ROOT_TYPE)) to '$(readvar MOUNTS_ROOT_TARGET)' with options: $(readvar MOUNTS_ROOT_OPTIONS)"
+    retry {self["mount_retries"] or -1} {self["mount_timeout"]} mount "$(readvar MOUNTS_ROOT_SOURCE)" -t "$(readvar MOUNTS_ROOT_TYPE)" "$(readvar MOUNTS_ROOT_TARGET)" -o "$(readvar MOUNTS_ROOT_OPTIONS)"
+    """
 
 
 def export_mount_info(self) -> None:
@@ -773,6 +815,9 @@ def export_mount_info(self) -> None:
 
 def autodetect_mount_kmods(self, device) -> None:
     """Autodetects the kernel modules for a block device."""
+    if "/" not in str(device):
+        device = f"/dev/{device}"
+
     if device_kmods := resolve_blkdev_kmod(self, device):
         self.logger.info("Auto-enabling kernel modules for device: %s" % colorize(", ".join(device_kmods), "cyan"))
         self["_kmod_auto"] = device_kmods
@@ -780,27 +825,40 @@ def autodetect_mount_kmods(self, device) -> None:
 
 def resolve_blkdev_kmod(self, device) -> list[str]:
     """Gets the kmod name for a block device."""
+    kmods = []
     dev = Path(device)
     while dev.is_symlink():
         dev = dev.resolve()
+
+    if dev.is_block_device():
+        major, minor = _get_device_id(device)
+        sys_dev = str(Path(f"/sys/dev/block/{major}:{minor}").resolve())
+        if "/usb" in sys_dev:
+            if "ugrd.kmod.usb" not in self["modules"]:
+                self.logger.info(
+                    "Auto-enabling %s for USB device: %s"
+                    % (colorize("ugrd.kmod.usb", bold=True), colorize(device, "cyan"))
+                )
+                self["modules"] = "ugrd.kmod.usb"
     device_name = dev.name
+
     if device_name.startswith("dm-") or dev.parent.name == "mapper" or dev.parent.name.startswith("vg"):
-        return ["dm_mod"]
+        kmods.append("dm_mod")
     elif device_name.startswith("nvme"):
-        return ["nvme"]
+        kmods.append("nvme")
     elif device_name.startswith("vd"):
-        return ["virtio_blk"]
+        kmods.append("virtio_blk")
     elif device_name.startswith("sd"):
-        return ["sd_mod"]
+        kmods.append("sd_mod")
     elif device_name.startswith("mmcblk"):
-        return ["mmc_block"]
+        kmods.append("mmc_block")
     elif device_name.startswith("sr"):
-        return ["sr_mod"]
+        kmods.append("sr_mod")
     elif device_name.startswith("md"):
-        return ["md_mod"]
+        kmods.append("md_mod")
     else:
         self.logger.error(
             "[%s] Unable to determine kernel module for block device: %s"
             % (device_name, colorize(device, "red", bold=True))
         )
-        return []
+    return kmods
